/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Public License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Public License, please send an email to 
 * dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Public License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;

using Microsoft.Scripting.Generation;
using Microsoft.Scripting.Runtime;
using Microsoft.Scripting.Utils;

namespace Microsoft.Scripting.Ast {
    /// <summary>
    /// The base class for LambdaBinder and RuleBinder.
    /// </summary>
    internal sealed class VariableBinder : Walker {
        // Information collected about the scopes while doing variable binding
        private sealed class ScopeBindingInfo {
            // ScopeInfo is created during post processing
            internal ScopeInfo ScopeInfo;

            // true if this scope closes over variables in a parent lambda
            internal bool IsClosure;

            // true if this scope contains a yield statement
            internal bool HasYield;

            internal readonly LambdaExpression Lambda;
            internal readonly Expression Scope;
            internal readonly ScopeBindingInfo Parent;
            internal readonly List<VariableBindingInfo> Variables = new List<VariableBindingInfo>();

            internal ScopeBindingInfo(Expression scope, LambdaExpression lambda, ScopeBindingInfo parent) {
                Scope = scope;
                Lambda = lambda;
                Parent = parent;
            }

            internal bool LiftAll {
                get {
                    return Lambda.NodeType == ExpressionType.Generator && (HasYield || Scope == Lambda);
                }
            }
        }

        // Information collected about the variables while doing variable binding
        private sealed class VariableBindingInfo {
            // true if this variable is lifted to an environment
            internal bool IsLifted;
            internal readonly Expression Variable;
            internal readonly ScopeBindingInfo Scope;

            internal VariableBindingInfo(Expression variable, ScopeBindingInfo scope) {
                Variable = variable;
                Scope = scope;
                scope.Variables.Add(this);
            }
        }

        // The list of all scopes, in prefix order, for post processing
        private readonly List<ScopeBindingInfo> _scopeList = new List<ScopeBindingInfo>();

        // The dictionary of all scopes and their infos in the tree
        private Dictionary<Expression, ScopeBindingInfo> _scopes = new Dictionary<Expression, ScopeBindingInfo>();

        // The dictionary of all variables and their infos in the tree
        private Dictionary<Expression, VariableBindingInfo> _variables = new Dictionary<Expression, VariableBindingInfo>();

        // The dictionary of all generators and their infos in the tree.
        private Dictionary<LambdaExpression, GeneratorInfo> _generators;

        // The ScopeInfos that are the result of the variable binder
        private Dictionary<Expression, ScopeInfo> _resultScopes;

        // Stack to keep track of scope nesting.
        private Stack<ScopeBindingInfo> _stack = new Stack<ScopeBindingInfo>();

        private VariableBinder() {
        }

        #region entry point

        /// <summary>
        /// LambdaBinder entry point.
        /// </summary>
        internal static AnalyzedTree Bind(LambdaExpression ast) {
            return new VariableBinder().BindLambda(ast);
        }

        #endregion

        private AnalyzedTree BindLambda(LambdaExpression lambda) {
            // Collect the lambdas
            WalkNode(lambda);
            BindTheScopes();

            return new AnalyzedTree(_resultScopes, _generators);
        }

        private LambdaExpression GetTopLambda() {
            return _stack.Peek().Lambda;
        }

        #region Walker overrides

        protected override bool Walk(VariableExpression node) {
            Reference(node);
            return true;
        }

        protected override bool Walk(ParameterExpression node) {
            Reference(node);
            return true;
        }

        protected override bool Walk(LambdaExpression node) {
            return Push(node);
        }

        protected override void PostWalk(LambdaExpression node) {
            if (node.NodeType == ExpressionType.Generator) {
                if (_generators == null) {
                    _generators = new Dictionary<LambdaExpression, GeneratorInfo>();
                }

                // the info may already exist if we've already processed this lambda
                // (i.e. it appears in multiple places in the tree)
                if (!_generators.ContainsKey(node)) {

                    GeneratorInfo gi = YieldLabelBuilder.BuildYieldTargets(node);
                    _generators.Add(node, gi);

                    // For scopes inside a generator, we need to lift the scope
                    // access slots. Otherwise, the scope would be null when
                    // calling back into the generator.
                    for (int i = _scopeList.Count - 1; i >= 0; i--) {
                        ScopeBindingInfo s = _scopeList[i];

                        // stop when we get to the generator itself
                        if (s.Scope == node) {
                            break;
                        }

                        // If this scope exists immediately in the generator and
                        // yields, we want to lift its access slot and use a
                        // generator temp instead of an IL local
                        if (s.Lambda == node && s.HasYield) {

                            VariableExpression temp = gi.AddGeneratorTemp();

                            // mark closures up to the generator
                            for (ScopeBindingInfo t = s; t.Scope != node; t = t.Parent) {
                                t.IsClosure = true;
                            }
                        }
                    }

                    foreach (VariableExpression v in gi.Temps) {
                        DefineVariable(v);
                    }
                }
            }

            ScopeBindingInfo sbi = Pop();
            Debug.Assert(sbi.Scope == node);
        }

        // This may not belong here because it is checking for the
        // AST type consistency. However, since it is the only check
        // it seems unwarranted to make an extra walk of the AST just
        // to verify this condition.
        protected override bool Walk(ReturnStatement node) {
            LambdaExpression lambda = GetTopLambda();
            Type returnType = lambda.ReturnType;

            if (node.Expression != null) {
                if (lambda.NodeType == ExpressionType.Lambda) {
                    if (!returnType.IsAssignableFrom(node.Expression.Type)) {
                        throw InvalidReturnStatement("Invalid type of return expression value", node);
                    }
                }
            } else {
                // return without expression can be only from lambda with void return type
                if (returnType != typeof(void)) {
                    throw InvalidReturnStatement("Missing return expression", node);
                }
            }

            return true;
        }

        private static ArgumentException InvalidReturnStatement(string message, ReturnStatement node) {
            SourceSpan span = node.Annotations.Get<SourceSpan>();
            return new ArgumentException(
                String.Format(
                    "{0} at {1}:{2}-{3}:{4}", message,
                    span.Start.Line, span.Start.Column, span.End.Line, span.End.Column
                )
            );
        }

        protected override bool Walk(ScopeExpression node) {
            return Push(node);
        }

        protected override void PostWalk(ScopeExpression node) {
            ScopeBindingInfo sbi = Pop();
            Debug.Assert(sbi.Scope == node);
        }

        protected override void PostWalk(YieldStatement node) {
            // Mark the whole scope chain in this generator as having a yield
            ScopeBindingInfo sbi = _stack.Peek();

            ScopeBindingInfo current = sbi;
            while (current.Scope.NodeType != ExpressionType.Generator) {
                Debug.Assert(current.Lambda == sbi.Lambda);

                current.HasYield = true;
                current = current.Parent;
            }
        }

        protected override void PostWalk(LocalScopeExpression node) {
            // Ensure that all variables are defined in an outer scope
            // Force lifting of variables if needed

            ScopeBindingInfo current = _stack.Peek();
            foreach (Expression v in node.Variables) {
                // Make sure the variable is defined
                VariableBindingInfo vbi;
                if (!_variables.TryGetValue(v, out vbi)) {
                    throw InvalidVariableReference(v);
                }

                // Close over the variable if necessary
                if (node.IsClosure) {
                    // reference it from this scope
                    Reference(v);

                    // force lifting even if it's in the same lambda
                    vbi.IsLifted = true;
                }
            }
        }

        #endregion

        #region processing scope expressions

        private bool Push(Expression scope) {
            // We've seen this scope already
            // (referenced from multiple places in the tree)
            ScopeBindingInfo sbi;
            if (_scopes.TryGetValue(scope, out sbi)) {
                // Push the expression so PostWalk can pop it
                _stack.Push(sbi);
                return false;
            }

            // The parent of the scope is the scope currently
            // on top of the stack, or a null if at top level.
            ScopeBindingInfo parent = null;
            LambdaExpression lambda = null;

            if (_stack.Count > 0) {
                parent = _stack.Peek();
                lambda = parent.Lambda;
            }

            if (scope.NodeType != ExpressionType.Scope) {
                lambda = (LambdaExpression)scope;
            }

            sbi = new ScopeBindingInfo(scope, lambda, parent);

            // Remember we saw the scope already
            _scopes[scope] = sbi;

            // Store it for post processing
            _scopeList.Add(sbi);

            // And push it on the stack.
            _stack.Push(sbi);

            // define variables in this scope
            DefineVariables(scope);

            // if the child of this scope is another scope, merge them
            return MergeWithChildScope(scope);
        }

        // Merges variables from child scopes into the parent
        // This is an optimization for the common patterm:
        //      Expression.Lambda(Expression.Scope(...), ...)
        //
        // It saves both environment allocations and speeds up access
        //
        private bool MergeWithChildScope(Expression scope) {

            Expression body;
            if (scope.NodeType == ExpressionType.Scope) {
                body = ((ScopeExpression)scope).Body;
            } else {
                body = ((LambdaExpression)scope).Body;
            }

            if (body.NodeType == ExpressionType.Scope) {

                // While the body is a ScopeExpression, merge the variables
                // into this scope
                do {
                    ScopeExpression se = (ScopeExpression)body;
                    foreach (VariableExpression v in se.Variables) {
                        DefineVariable(v);
                    }
                    body = se.Body;

                } while (body.NodeType == ExpressionType.Scope);

                // we finally found a non-scope body, walk it
                WalkNode(body);

                // return false so we don't walk the body again
                return false;
            }

            // continue normal walking of the tree
            return true;
        }

        private void DefineVariables(Expression scope) {
            if (scope.NodeType == ExpressionType.Scope) {
                ScopeExpression se = (ScopeExpression)scope;
                foreach (VariableExpression v in se.Variables) {
                    DefineVariable(v);
                }
            } else {
                LambdaExpression le = (LambdaExpression)scope;
                foreach (ParameterExpression p in le.Parameters) {
                    DefineVariable(p);
                }
            }
        }

        private void DefineVariable(Expression v) {
            if (_variables.ContainsKey(v)) {
                throw InvalidVariableDefinition(v);
            }

            _variables.Add(v, new VariableBindingInfo(v, _stack.Peek()));
        }

        private ScopeBindingInfo Pop() {
            Debug.Assert(_stack != null && _stack.Count > 0);
            ScopeBindingInfo sbi = _stack.Pop();

            // Remove variables that are no longer in scope
            // (this allows other parallel scopes to reuse variable instances)
            foreach (VariableBindingInfo vbi in sbi.Variables) {
                _variables.Remove(vbi.Variable);
            }

            return sbi;
        }

        #endregion

        #region Closure resolution

        /// <summary>
        /// Called when a variable is referenced inside the current lambda.
        /// 
        /// If the variable is defined on an outer lambda:
        ///   1. Mark the current lambda and lambdas between as closures
        ///   2. Mark the variable as lifted
        /// </summary>
        private void Reference(Expression variable) {
            Debug.Assert(variable != null && _stack != null && _stack.Count > 0);

            VariableBindingInfo vbi;
            if (!_variables.TryGetValue(variable, out vbi)) {
                throw InvalidVariableReference(variable);
            }

            ScopeBindingInfo referenceScope = _stack.Peek();
            ScopeBindingInfo definingScope = vbi.Scope;

            // Mark all scopes between the use and the definition as closures
            // If the variable ends up getting lifted, we need to be able to
            // get at its environment.
            for (ScopeBindingInfo s = referenceScope; s != definingScope; s = s.Parent) {
                s.IsClosure = true;
            }

            // If it's defined in this lambda, don't lift it.
            // Only lift variables that are closed over, or need to be lifted
            // for some other reason (generators, scope access)
            if (definingScope.Lambda != referenceScope.Lambda) {
                vbi.IsLifted = true;
            }
        }

        private Exception InvalidVariableReference(Expression variable) {
            return new InvalidOperationException(
                string.Format(
                    "Variable '{0}' referenced from scope '{1}', but is is not defined in an outer scope",
                    CompilerHelpers.GetVariableName(variable),
                    ScopeInfo.GetName(_stack.Peek().Scope) ?? "<unnamed>"
                )
            );
        }

        private Exception InvalidVariableDefinition(Expression variable) {
            return new InvalidOperationException(
                string.Format(
                    "Variable '{0}' definined in scope '{1}', but it is already defined in '{2}'",
                    CompilerHelpers.GetVariableName(variable),
                    ScopeInfo.GetName(_stack.Peek().Scope) ?? "<unnamed>",
                    ScopeInfo.GetName(_variables[variable].Scope.Scope) ?? "<unnamed>"
                )
            );
        }

        /// <summary>
        /// Post processing of the tree:
        ///   1. Lift all locals to environment if necessary (generators,
        ///      lambdas marked with EmitLocalDictionary, etc)
        ///   2. Group variables into lifted, locals, and globals
        ///   3. Finally, create a ScopeInfos for each scope we encountered
        /// </summary>
        private void BindTheScopes() {
            Debug.Assert(_stack.Count == 0);

            // free items we don't need anymore
            _variables = null;
            _scopes = null;
            _stack = null;

            // allocate the result dictionary
            _resultScopes = new Dictionary<Expression, ScopeInfo>(_scopeList.Count);

            // walk each scope, and build the ScopeInfo
            foreach (ScopeBindingInfo sbi in _scopeList) {
                Debug.Assert(sbi.ScopeInfo == null);

                BindScope(sbi);                
            }
        }

        private void BindScope(ScopeBindingInfo sbi) {
            List<Expression> lifted = new List<Expression>();
            List<Expression> locals = new List<Expression>();

            foreach (VariableBindingInfo vbi in sbi.Variables) {
                if (vbi.IsLifted || vbi.Scope.LiftAll) {
                    lifted.Add(vbi.Variable);
                } else {
                    locals.Add(vbi.Variable);
                }
            }

            ScopeInfo parent = null;
            if (sbi.Parent != null) {
                parent = sbi.Parent.ScopeInfo;
                Debug.Assert(parent != null);
            }

            ScopeInfo si = new ScopeInfo(
                parent,
                sbi.Scope,
                sbi.IsClosure,
                sbi.HasYield,
                new ReadOnlyCollection<Expression>(lifted.ToArray()),
                new ReadOnlyCollection<Expression>(locals.ToArray())
            );

            // Create the scope for the inner generator.
            //
            // The inner generator scope doesn't have any variables, but closes
            // over all variables from the outer generator method.
            //
            // We do this here so we don't have to play games with state inside
            // ScopeInfo, by ensuring that the scope chain mirrors the actual
            // generated code.
            if (si.Expression.NodeType == ExpressionType.Generator) {
                si = new ScopeInfo(
                    si,
                    sbi.Scope,
                    true,  // isClosure
                    false, // hasYield
                    DefaultReadOnlyCollection<Expression>.Empty,
                    DefaultReadOnlyCollection<Expression>.Empty
                );
            }

            _resultScopes.Add(si.Expression, sbi.ScopeInfo = si);
        }

        #endregion
    }
}
