/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Public License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Public License, please send an email to 
 * dlr@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Public License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using Microsoft.Contracts;
using Microsoft.Scripting.Utils;

namespace Microsoft.Scripting.Generation {
    /// <summary>
    /// Slot refers to a reference to an object. For eg, a global variable, a local variable, etc.
    /// A Slot is referred to using a Name. The Namespace is used to map a Name to a Slot.
    /// Multiple Names can refer to the same Slot.
    /// For eg. multiple closures can refer to the same Slot of a local variable in the enclosing
    /// function. Though each closure will use the same string (the name of the variable), each
    /// string is considered a unique Name or symbol.
    /// </summary>
    abstract class Slot {
        public abstract void EmitGet(ILGen cg);
        public abstract void EmitGetAddr(ILGen cg);

        #region emit Set methods
        // Must override at least one of these two methods or get infinite loop
        // 
        public virtual void EmitSet(ILGen cg, Slot val) {
            ContractUtils.RequiresNotNull(val, "val");
            ContractUtils.RequiresNotNull(cg, "cg");

            val.EmitGet(cg);
            EmitSet(cg);
        }

        // This override assumes that the IL stack already holds the value to be assigned from.
        // This default implementation assumes the top of stack is the slot type.
        public virtual void EmitSet(ILGen cg) {
            ContractUtils.RequiresNotNull(cg, "cg");

            // localTmpVal = <top of IL stack>
            Slot localTmpVal = cg.GetLocalTmp(Type);
            localTmpVal.EmitSet(cg);

            // <slot> = localTmpVal
            EmitSet(cg, localTmpVal);

            cg.FreeLocalTmp(localTmpVal);
        }
        #endregion // emit Set methods

        public abstract Type Type { get; }

        [Confined]
        public override string/*!*/ ToString() {
            return String.Format("{0} Type: {1}", GetType().Name, Type.FullName);
        }
    }
}
